/* Generated by Cython 0.24 */

#ifndef __PYX_HAVE__pachi_py__cypachi
#define __PYX_HAVE__pachi_py__cypachi


#ifndef __PYX_HAVE_API__pachi_py__cypachi

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#ifndef DL_IMPORT
  #define DL_IMPORT(_T) _T
#endif

__PYX_EXTERN_C DL_IMPORT(PyObject) *_PyIllegalMove;
__PYX_EXTERN_C DL_IMPORT(PyObject) *_PyPachiEngineError;

#endif /* !__PYX_HAVE_API__pachi_py__cypachi */

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC initcypachi(void);
#else
PyMODINIT_FUNC PyInit_cypachi(void);
#endif

#endif /* !__PYX_HAVE__pachi_py__cypachi */
